#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import string as _string
import math
#import random
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message




############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################

# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>>Custom Font Tools>>Custom Font Faux 3D v. 0.02
# Version 0.01 Options for optionally adding a decorative layer and creating a 3D style effect using multiple layers with a custom font file
#         Depth, direction of 3D and shading can be modified through the filter dialog 
#         Option for modifying existing kerning
# Version 0.02 New style added with a slightly more perspective view of the 3d layering using scaling
#         Modes to darken 3D layers, blurring and shading values of the 3D layers now available in the dialog
# Version 0.03 Adds preview options to display a character with the current values in a new image
# Version 0.04 Adds Tabbed dialog (Arakne & MareroQ) and options for creating a separate image using designated text with perspective options including scaling and depth


#from gimpfu import *
#from array import array
#import math


# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
debug_string = ""

def do_stuff(image, layer, style, scale_inc, depth, blur, use_decor, initial_shade, shade_range, darken_mode, direction, kerning_adjustment, do_perspective, perspective_word, persp_spacing, persp_vertical_justify, persp_horizontal_justify, scaled_depth, scaled_scale, to_visible, do_preview, preview_char, ptab):
    #### Main Filter Proc

    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
	
    num_layers, layer_ids = pdb.gimp_image_get_layers(image)
    layer_names_list = image.layers
    visi_loop, custom_check = 0, False
    while visi_loop < num_layers:
        # get info about layer we are working on
        layer_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
        char_layer_name = layer_name.decode('utf-8')
        # check to see if the current layer is a custom font character layer
        test_for_custom_layer = fn_test_for_custom_layer(image, layer_names_list[visi_loop], char_layer_name)
			
        if test_for_custom_layer:
            custom_check = True
            visi_loop = num_layers
			
        visi_loop += 1
		
    debug_string = "Perspective: UI depth = " + str(depth) + " : " + " UI Direction = " + str(direction) + " : "
    if do_perspective: # add 3d effect to range of characters only
        custom_check, do_preview, use_decor = False, False, False # Force unwanted options off
        debug_string = "Perspective: UI depth = " + str(depth) + " : " + " UI Direction = " + str(direction) + " : "
		
        loop, test_exists, new_word = 0, True, ""
        while loop<len(perspective_word):
            test_layer = perspective_word[loop]
            test_char = test_layer
            if test_layer == " ": 
                test_layer = "space"
                test_char = " "
                #new_word += test_char
            test_layer = test_layer + ".png"
            check_layer = pdb.gimp_image_get_layer_by_name(image, test_layer)
            if check_layer<0:
                test_char = test_layer[0]
                if test_char.isupper():
                    test_char = test_char.lower()
                    test_layer = test_char + ".png"
                    check_layer = pdb.gimp_image_get_layer_by_name(image, test_layer)
                    if check_layer<0: 
                        test_exists = False
                    else:
                        new_word += test_char
                elif test_char.islower():
                    test_char = test_char.upper()
                    test_layer = test_char + ".png"
                    check_layer = pdb.gimp_image_get_layer_by_name(image, test_layer)
                    if check_layer<0: 
                        test_exists = False
                    else:
                        new_word +=  test_char
            else:
                new_word +=  test_char
            loop +=1
        #pdb.gimp_message(" test_exists = " + str(test_exists) + " new_word = " + str(new_word)) 
        perspective_word = new_word
        if test_exists == False: do_perspective = False
    else:
        persp_step, has_centre, left_perspective, right_perspective, persp_space_width, scaled_depth, scaled_scale = 0, False, 0, 0, 0, 0, 0
		
            
	
    if custom_check: # add 3d effect to a custom font file
	
        test_for_custom_layer = False
		
        max_width, im_height, debug_string = proc_do_3D(image, style, scale_inc, depth, blur, use_decor, initial_shade, shade_range, darken_mode, direction, kerning_adjustment, do_preview, preview_char, do_perspective, persp_step, has_centre, left_perspective, right_perspective, persp_space_width, perspective_word, scaled_depth, scaled_scale, debug_string)
	
        if not do_preview: pdb.gimp_image_resize(image, max_width, im_height, 0, 0) # resize canvas to fit characters
	
        if (kerning_adjustment<> 0) and (do_preview == False): # adjust kerning paths if required by dialog
			
        # get lists of vector ids in image
            num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
            vector_list = image.vectors
            loop = 0
            while loop<num_vectors:
		
                # look for recognised kerning path
                kerning_path_name = pdb.gimp_item_get_name(vector_list[loop])
                kerning_path_name = kerning_path_name.decode('utf-8')
			
                if (len(kerning_path_name) == 5 and ".png" in kerning_path_name) or (len(kerning_path_name) == 6 and ".png" in kerning_path_name):
                    
                    kerning_vector = vector_list[loop]
                    num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(kerning_vector)
                    type, num_points, controlpoints, closed  = pdb.gimp_vectors_stroke_get_points(kerning_vector, stroke_ids[0])
                    kerning_target_width = controlpoints[6] + depth
                    v_position = pdb.gimp_image_get_vectors_position(image, kerning_vector)
                    pdb.gimp_image_remove_vectors(image, kerning_vector)
                    new_vector_id = pdb.gimp_vectors_new(image, kerning_path_name)
                    pdb.gimp_image_insert_vectors(image, new_vector_id, None, v_position)
                    points1 = [controlpoints[0],controlpoints[1]] * 3
                    points2 = [kerning_target_width,controlpoints[7]] * 3
                    points = points1 + points2
                    new_vector_id = pdb.gimp_vectors_stroke_new_from_points(new_vector_id ,0, len(points), points, False)
                loop += 1
    else:
        if do_perspective:
		
		
            persp_space_layer = pdb.gimp_image_get_layer_by_name(image, "space.png") # get width of space character
            #pdb.gimp_message( "persp_space_layer = " + str(persp_space_layer))
            if persp_space_layer == -1 or persp_space_layer == None:
                persp_space_width = image.width
                persp_space_height = image.height
            else:
                persp_space_width = persp_space_layer.width
                persp_space_height = persp_space_layer.height
			
            if len(perspective_word)<2: perspective_word = "GIMP" # sanity check so we have something to work on
            original_depth = depth
            #if style == 1 and scaled_depth<0:
            if scaled_scale<0:
                if depth - (abs(scaled_depth)*(len(perspective_word)-1) )<4: depth = (abs(scaled_depth)*(len(perspective_word)-1) + 4) # ensure that depth setting will work for negative scaling
			
            has_centre = True # identify central character position in perspective word
            if len(perspective_word) % 2 == 0: has_centre = False
            left_perspective = (len(perspective_word)//2)-1
            if has_centre:
                right_perspective = left_perspective + 2
            else:
                right_perspective = left_perspective + 1
				
            persp_step = 0
            if style == 0:
                persp_step = depth//(len(perspective_word) // 2) # get step size for changes in perspective depth
                if persp_step<4: 
                    persp_step = 4
                    depth = (len(perspective_word) // 2) * 4
			
            debug_string += "Perspective step = " + str(persp_step) +": Adjusted depth = " + str(depth) + ": " + "left_perspective = " + str(left_perspective) + ": "
			
            
            #pdb.gimp_message("depth = " + str(depth))
			
            perspective_image = pdb.gimp_image_new(image.width, image.height, 0) # add a new image for the perspective word
            do_autocrop = False
            if perspective_word.isupper(): do_autocrop=True # no cropping to layer content if using lower-case
            loop, persp_min_width = 0, 100000000
            num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
            vector_list = image.vectors
			
			
            while loop < len(perspective_word): # copy the characters from the custom font image file to the perspective image
                if str(perspective_word[loop]) == " ":
                    perspective_layer_name = "space.png"
                else:
                    perspective_layer_name = str(perspective_word[loop])+".png"
					
                temp_layer_id = pdb.gimp_image_get_layer_by_name(image, perspective_layer_name)
                perspective_layer_id = pdb.gimp_layer_new_from_drawable(temp_layer_id, perspective_image)
                pdb.gimp_image_insert_layer(perspective_image, perspective_layer_id, None, loop) # ensure layer order is the same as in the perspective word
                pdb.gimp_item_set_visible(perspective_layer_id, True)
                temp_layer_copy = perspective_layer_id.copy(False)
                perspective_image.add_layer(temp_layer_copy,0)
				
                # note original width for future scaling of kerning paths 
                if loop == 0:
                    original_width_list = [temp_layer_copy.width]
                    #pdb.gimp_message("original_width = " + str(temp_layer_id.width) )
                else:
                    original_width_list.append(temp_layer_copy.width)
                    #pdb.gimp_message("original_width = " + str(temp_layer_id.width) )
                
                pdb.gimp_image_set_active_layer(perspective_image, temp_layer_copy)
                if do_autocrop: pdb.plug_in_autocrop_layer(perspective_image, temp_layer_copy) # crop characters to content if no lower case
                if temp_layer_copy.width<persp_min_width: persp_min_width = temp_layer_copy.width # needed for scaling sanity check
                if loop == 0:
                    persp_size = [temp_layer_copy.width]
                    if do_autocrop:
                        left_margin_list = [(original_width_list[loop]-temp_layer_copy.width)/2] # attempt to estimate transparent left/right margins for kerning adjustment
                    else:
                        pdb.gimp_image_select_item(perspective_image, 2, temp_layer_copy)
                        non_empty, sx1, sy1, sx2, sy2  = pdb.gimp_selection_bounds(perspective_image)
                        pdb.gimp_selection_none(perspective_image)
                        if non_empty:
                            left_margin_list = [sx1]
                        else:
                            left_margin_list = [0]
						
                else:
                    persp_size.append(temp_layer_copy.width)
                    if do_autocrop:
                        left_margin_list.append((original_width_list[loop]-temp_layer_copy.width)/2)
                        #pdb.gimp_message("left_margin_list = " + str((original_width_list[loop]-temp_layer_copy.width)/2) + " original_width_list[loop] = " + str(original_width_list[loop]) + " temp_layer_copy.width = " + str(temp_layer_copy.width) )
                    else:
                        pdb.gimp_image_select_item(perspective_image, 2, temp_layer_copy)
                        non_empty, sx1, sy1, sx2, sy2  = pdb.gimp_selection_bounds(perspective_image)
                        pdb.gimp_selection_none(perspective_image)
                        if non_empty:
                            left_margin_list.append(sx1)
                            #pdb.gimp_message("sx1 = " + str(sx1) )
                        else:
                            left_margin_list.append(0)
                if loop < len(perspective_word)-1:
                    kerning_pair = perspective_word[loop] + perspective_word[loop + 1] +".png" # create a kerning path name
                    kern_vector = pdb.gimp_image_get_vectors_by_name(image,str(kerning_pair) ) # and see if it exists in the parent image
                    
                    if kern_vector<0:
                        kerning_value = 0 # ie no kerning path available
                    else:
                        NumberOfStrokes_in_Path,ActivePath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(kern_vector) # get vector information for kerning path width
                        ActivePath_StrokeID = ActivePath_StrokeIDs_List[0]
                        kerning_value = pdb.gimp_vectors_stroke_get_length(kern_vector,ActivePath_StrokeID,0.01) # ie the length of the existing kerning path
                        
                    if loop == 0: # add kerning path lengths to a list
                    
                        kerning_list = [kerning_value]
                    else:
                    
					
                        kerning_list.append(kerning_value)
                perspective_image.remove_layer(temp_layer_copy) # remove temporary layer
                loop += 1
				
				
            perspective_display = pdb.gimp_display_new(perspective_image) # display the perspective word characters
			
            num_persp_layers, persp_layer_ids = pdb.gimp_image_get_layers(perspective_image) # get info on perspective image layers
            persp_layer_names_list = perspective_image.layers
			
            #pdb.gimp_message(str(persp_layer_names_list))
            persp_depth = depth
			
            max_width, im_height, debug_string = proc_do_3D(perspective_image, style, scale_inc, depth, blur, use_decor, initial_shade, shade_range, darken_mode, direction, kerning_adjustment, do_preview, preview_char, do_perspective, persp_step, has_centre, left_perspective, right_perspective, persp_space_width, perspective_word, scaled_depth, scaled_scale, debug_string)
			
			
            # position characters for word
            num_persp_layers, persp_layer_ids = pdb.gimp_image_get_layers(perspective_image) # get info on perspective image layers
            persp_layer_names_list = perspective_image.layers
			
            if style == 0: # Translated style
                word_x, word_y = 0, 0
                loop, offset = 0, depth
                while loop < len(perspective_word):
                    if loop == 0:
                        offset_list=[word_y, offset]
                    else:
                        offset_list.append(word_y)
                        offset_list.append(offset)
                    temp_layer = persp_layer_names_list[loop]
                    pdb.gimp_image_set_active_layer(perspective_image, temp_layer) # crop characters to content
                    if do_autocrop: pdb.plug_in_autocrop_layer(perspective_image, temp_layer)
                    temp_layer.set_offsets(0, 0)
				
                    if direction ==4 or direction == 0: # vertical
                    
                        if persp_vertical_justify == 1: # centre justify
                            word_x=(image.width - temp_layer.width)/2
                        elif persp_vertical_justify == 2: # right justify
                            word_x=(image.width - temp_layer.width)
                    
						
                        if loop<left_perspective:
                            offset -= persp_step
                        elif loop>left_perspective+1:
                            offset += persp_step
                        word_y -= offset
					
                        word_y += temp_layer.height
                        pdb.gimp_item_transform_translate(temp_layer, word_x, word_y)
                        pdb.gimp_item_set_visible(temp_layer, True)
					
					
				
                    else:
                        
                        if persp_horizontal_justify == 2: # justify to bottom
                            word_y = persp_layer_names_list[0].height - temp_layer.height
                        elif persp_horizontal_justify == 1: # centred
                            word_y = (persp_layer_names_list[0].height - temp_layer.height)/2
							
                        pdb.gimp_item_transform_translate(temp_layer, word_x, word_y)
                        pdb.gimp_item_set_visible(temp_layer, True)
						
						
                        #scaled_kerning = temp_layer.width
						
                        scaled_kerning = persp_size[loop]
						
                        if loop<len(perspective_word)-1: 
                            if (kerning_list[loop]) > 0: scaled_kerning = ((float(kerning_list[loop])/(float(original_width_list[loop]))* persp_size[loop])) # scale the original kerning path length
                            #pdb.gimp_message("kerning = " + str(kerning_list[loop]) + " original width = " + str(original_width_list[loop]) + " scaled_kerning =  " + str(scaled_kerning) + " temp_layer.width = " + str(temp_layer.width) + " persp_size = " + str(persp_size[loop]) + " left_margin_list[loop] = " + str(left_margin_list[loop]) + " right_perspective = " + str(right_perspective) + " left_perspective = " + str(left_perspective) + " word_x = " + str(word_x) )
							
							
                        if direction == 1 or direction == 2 or direction == 3: # NE or SE
						
                            if not has_centre:
                                if loop == left_perspective:
                                    #word_x += temp_layer.width
                                    word_x += scaled_kerning
                                    word_x +=  depth ####
                                    if not do_autocrop: word_x -= (left_margin_list[loop])
					
                                else:
							
                                    word_x += scaled_kerning 
                                    word_x += depth/2
									
                                    
                                    if not do_autocrop: word_x -= (left_margin_list[loop])
							
                                    if loop>=right_perspective:  
                                        word_x-= (depth)
										
                            else:
							
                                word_x += scaled_kerning 
                                word_x += depth/2
								
                                if loop>=right_perspective-1:  
                                    word_x-= (depth)
                                    if not do_autocrop: word_x-= (left_margin_list[loop])
                                    
							
                        elif direction == 5 or direction == 6 or direction == 7: # SW or NW
						
                            if not has_centre:
							
                                word_x += scaled_kerning
                                if not do_autocrop: word_x-= (left_margin_list[loop])
								
                                if loop <= left_perspective: 
                                    word_x += (depth)
										
					
                            else:
							
                                word_x += scaled_kerning
                                if not do_autocrop: word_x-= (left_margin_list[loop])
								
                                if loop <= left_perspective: 
                                    word_x += (depth)
									
										
                        word_x += persp_spacing
						
						
                    loop += 1
                pdb.gimp_image_resize_to_layers(perspective_image)
			
			
                #re-order layers in layer stack
			
                persp_layer_names_list = perspective_image.layers
                if direction == 5 or direction == 7: # SW -> SE, NW -> NE
                    limit, loop = left_perspective, 0
                    while loop<=(left_perspective):
                        persp_layer_names_list = perspective_image.layers
                        i_loop = limit
                        while i_loop>=0:
                            pdb.gimp_image_raise_item(perspective_image, persp_layer_names_list[len(perspective_word)-1])
                            i_loop -= 1
                        loop += 1
                        limit -= 1
			
                elif direction == 1 or direction == 2 or direction == 3: #: SE -> SW, NE -> NW
			
                    limit, loop = 1, 1
                    while loop<(right_perspective):
                        persp_layer_names_list = perspective_image.layers
                        i_loop = 0
                        while i_loop<limit:
                            pdb.gimp_image_raise_item(perspective_image, persp_layer_names_list[loop])
                            i_loop += 1
                        loop += 1
                        limit += 1
				
                elif direction == 4: # S -> N
                    limit, loop = 1, left_perspective
                    while loop>-1:
                        persp_layer_names_list = perspective_image.layers
                        i_loop = 0
                        while i_loop<limit:
                            pdb.gimp_image_lower_item(perspective_image, persp_layer_names_list[loop])
                            i_loop += 1
                        limit+= 1
                        loop -= 1
						
            else: # scaled style		
			
                if scaled_scale<0:
                    min_width, min_height = 0, 0
                    while min_width<50 or min_height<50: # prevent negative scaling from reducing characters too much
                        loop = 0
                        while loop < len(perspective_word): # remove transparent borders from characters

                            temp_layer = persp_layer_names_list[loop]
                            pdb.gimp_image_set_active_layer(perspective_image, temp_layer)
                            if "space" in pdb.gimp_item_get_name(temp_layer) and do_autocrop: # if this word has a space character perform crop of space layer to reduce height
                                pdb.gimp_layer_resize(temp_layer, persp_space_width, persp_space_height/2, 0, 0)
                            else:
                                if do_autocrop: pdb.plug_in_autocrop_layer(perspective_image, temp_layer)
                            temp_layer.set_offsets(0, 0)
                            if loop == 0:
                                reduced_width_list = [temp_layer.width] ######## check that we are still using this list
                                reduced_height_list = [temp_layer.height] ######## check that we are still using this list
                                scaled_scale_list = [scaled_scale*loop] ######## check that we are still using this list
                                scale_factor_list = [0] ######## check that we are still using this list
                            
                            else:
                                if scaled_scale==0:
                                    reduced_width_list.append(temp_layer.width)
                                    reduced_height_list.append(temp_layer.height)
                                else:
                                    reduced_width_list.append(((float(scaled_scale*loop))*(float(float(temp_layer.width)/float(temp_layer.height) ))) + temp_layer.width)
                                    reduced_height_list.append(temp_layer.height + (float(float(temp_layer.width)/float(temp_layer.height) )) )
							
							
                            
                            loop += 1
                        min_width = min(reduced_width_list)
                        min_height = min(reduced_height_list)
                        if min_width<50 or min_height<50: 
                            scaled_scale += 10
                            if scaled_scale>0: scaled_scale = 0
					
                else:
                    loop = 0
                    while loop < len(perspective_word): # 

                        temp_layer = persp_layer_names_list[loop]
                        pdb.gimp_image_set_active_layer(perspective_image, temp_layer)
                        if "space" in pdb.gimp_item_get_name(temp_layer) and do_autocrop: # crop space character if required
                            pdb.gimp_layer_resize(temp_layer, persp_space_width, persp_space_height/2, 0, 0)
                        else:
                            if do_autocrop: pdb.plug_in_autocrop_layer(perspective_image, temp_layer) # crop to layer content if no lower case in perspective word
                        temp_layer.set_offsets(0, 0)
                        loop += 1
					
					
                #debug_string += " : reduced_width_list = " + str(reduced_width_list)
                #debug_string += " : reduced_height_list = " + str(reduced_height_list)
			
                word_x, word_y = 0, 0
                #loop, offset = 0, depth
                loop, offset = 0, original_depth
                scale_x, scale_y = scaled_scale, scaled_scale
				
                
                counter, depth = 0, original_depth
                uncropped_y = ((float(perspective_image.height+scale_x)/perspective_image.height)*scale_x)/4
                while loop < len(perspective_word):
				
                    temp_layer = persp_layer_names_list[loop]
                    pdb.gimp_image_set_active_layer(perspective_image, temp_layer) # crop characters to content
				
                        
                    if scaled_scale<>0 and loop > 0: # do not apply scaling to first character
                        persp_scale_factor, persp_width, persp_height = 0.0, float(temp_layer.width), float(temp_layer.height)
                        persp_scale_factor=float(persp_width/persp_height)
                        if scaled_scale<0:
                            pdb.gimp_layer_scale(temp_layer, temp_layer.width*(float(temp_layer.height+(float(scaled_scale)))/  temp_layer.height), temp_layer.height+(float((scaled_scale))), False)# scale by height rather than by width
                        if scaled_scale>0:
                            pdb.gimp_layer_scale(temp_layer, temp_layer.width*(float(temp_layer.height+(float(scaled_scale)))/  temp_layer.height), temp_layer.height+(float((scaled_scale))), False)# scale by height rather than by width
                        if scaled_scale<>0:
                            scaled_scale += scale_x
							
                            if direction == 0 or direction == 4: # North or South
                                dummy = 0
                            else: # apply horizontal justification			
                                if persp_horizontal_justify == 2: # justify to bottom
                                    word_y = persp_layer_names_list[0].height - temp_layer.height
                                elif persp_horizontal_justify == 1: # centred
                                    word_y = (persp_layer_names_list[0].height - temp_layer.height)/2
                                if not do_autocrop and scaled_scale<0: # y adjustment for uncropped characters
                                    if direction == 1 or direction == 3 or direction == 5 or direction == 7 or direction == 2 or direction == 6: word_y += (loop*uncropped_y)
                                #pdb.gimp_message("loop*uncropped_y = " + str(loop*uncropped_y) + " word_y = " + str(word_y) )
							
							
                    pdb.gimp_item_transform_translate(temp_layer, word_x, word_y)
                    pdb.gimp_item_set_visible(temp_layer, True)
						
						
                    if direction == 1 or direction == 3 or direction == 5 or direction == 7 or direction == 2 or direction == 6: # NE or SE or NW or SW or E or W
					

                        scaled_kerning = temp_layer.width # set char width
						
                        if scaled_scale==0:
                            if loop<len(perspective_word)-1: 
                                if kerning_list[loop] >0: 
                                    if do_autocrop:
                                        scaled_kerning = ((float(kerning_list[loop])/(float(original_width_list[loop])))* (temp_layer.width+left_margin_list[loop])) # Try B - slightly reduces scaled kerning amount compared to original kerning
                                        #pdb.gimp_message(" actual kerning = " + str(kerning_list[loop]) + " scaled_kerning = " + str(scaled_kerning) + " original_width = " + str(original_width_list[loop]) + " current width = " + str(temp_layer.width) + " left_margin_list[loop] = " + str(left_margin_list[loop]) + " word_x = " + str(word_x) )
                                    
                                    else:
										
                                        scaled_kerning = (kerning_list[loop]) # for uncropped width
										
                                else:
									
                                    if not do_autocrop: scaled_kerning-=(left_margin_list[loop])
						
							
                            if direction == 5 or direction == 7 or direction == 6: # NW or SW or W
                                if scaled_scale== 0:
                                    #word_x +=(left_margin_list[loop])
                                    word_x -= (depth) # adjust spacing to account for extra width of 3D chars with this direction
									
                            word_x += scaled_kerning # update current x position
							
							
                            
                        if scaled_scale>0:
                            #pdb.gimp_message("depth = " + str(depth))
                            if loop<len(perspective_word)-1: 
							
                                if kerning_list[loop] >0: 
                                    if do_autocrop:
                                        if persp_horizontal_justify == 1: # centred
                                            scaled_kerning = ((float(kerning_list[loop])/(float(original_width_list[loop])-((left_margin_list[loop])*2)))* (temp_layer.width-scale_x))
                                        else:
                                            scaled_kerning = ((float(kerning_list[loop])/(float(original_width_list[loop])-(((left_margin_list[loop])/2))))* (temp_layer.width))
                                    else:
                                        dummy = 0
                                        if loop == 0:
                                            scaled_kerning = ((float(kerning_list[loop])/(float(original_width_list[loop])+(left_margin_list[loop])))* (temp_layer.width))
                                        else:
                                            scaled_kerning = ((float(kerning_list[loop])/(float(original_width_list[loop])-(left_margin_list[loop])))* (temp_layer.width+scale_x))
								
								
                            if direction == 5 or direction == 7 or direction == 6: # NW or SW or W 
                                word_x -= (depth*2)
								
								
                            word_x += scaled_kerning # update current x position for increased scaling
								
                            
                        if scaled_scale<0:
						
                            if loop<len(perspective_word)-1: 
							
                                if kerning_list[loop] >0:  
                                    if do_autocrop:
                                        scaled_kerning = float(kerning_list[loop]/original_width_list[loop])*temp_layer.width
                                        word_x -= (((float(left_margin_list[loop]))/(scaled_scale))*(left_margin_list[loop])*loop) # adjusted spacing 
										
                                    else:
                                        scaled_kerning = (float(kerning_list[loop]/original_width_list[loop]))*temp_layer.width
                                        word_x += (((float(left_margin_list[loop]))/(scaled_scale))*(left_margin_list[loop])*loop) # adjusted spacing 
                                else: # adjust spacing for non-kerned characters
								
								
                                    if not do_autocrop:
                                        if loop>0:
                                            dummy = 0
                                            word_x += (((float(left_margin_list[loop]))/(scaled_scale))*(left_margin_list[loop])*loop)
                                        else:
                                            dummy = 0
                                            word_x += (((float(left_margin_list[loop]))/(scaled_scale))*(left_margin_list[loop])*loop)
                                    else:
                                        dummy = 0
                                        word_x -= (((float(left_margin_list[loop]))/(scaled_scale))*(left_margin_list[loop])*loop)
                                        
									
                                #pdb.gimp_message(" kerning_list = " + str(kerning_list) + " actual kerning = " + str(kerning_list[loop]) + " scaled_kerning = " + str(scaled_kerning) + " original_width = " + str(original_width_list[loop]) + " current width = " + str(temp_layer.width) + " left_margin_list[loop] = " + str(left_margin_list[loop]) + " word_x = " + str(word_x)  + " scaling = " + str(scale_x * loop))
                            word_x += scaled_kerning
							
                            if direction == 5 or direction == 7 or direction == 6: # NW or SW or W 
                                dummy = 0
                                word_x -=  ((depth*2/float(original_width_list[loop]))*scaled_kerning)
								
                        
                        word_x += persp_spacing
					
					
						
                    if direction == 0 or direction == 4: # N or S vertical positioning
                    #if direction == 0: # N or S vertical positioning
                        if scaled_scale<>0:
                            if scaled_scale>0: word_y += temp_layer.height/3 # 3d letter height
						
                        else:
                            word_y += temp_layer.height # 3d letter height
                        word_y += persp_spacing
						
                    depth += scaled_depth
                    loop += 1
					
                #if scaled_scale<0: pdb.gimp_message(str(temp_info_list) )
					
                if direction == 1 or direction == 3: # NE or SE
                    if scaled_scale>=0: # scaling not applied or increases letter sizes re-order characters front to back
                        loop = 1
                        while loop < len(perspective_word):
                            pdb.gimp_image_raise_item_to_top(perspective_image, persp_layer_names_list[loop])
                            loop += 1
							
							
							
                if direction == 0: # N
                    if scaled_scale<0:
                        loop, persp_y = len(perspective_word)-2, 0
                        while loop > -1:
                            temp_layer = persp_layer_names_list[loop]
                            persp_y += ((persp_layer_names_list[loop+1].height)/3)*2
                            pdb.gimp_item_transform_translate(temp_layer, 0, persp_y)
                            loop = loop - 1
                			
                if direction == 4: # S
                    if scaled_scale<0:
                        loop, persp_y = 1 , 0
                        while loop < len(perspective_word) :
                            temp_layer = persp_layer_names_list[loop]
                            persp_y += ((persp_layer_names_list[loop-1].height)/3)*2
                            #persp_y += ((persp_layer_names_list[loop-1].height)/2)
                            pdb.gimp_item_transform_translate(temp_layer, 0, persp_y)
                            loop = loop + 1
                
                if direction == 0 or direction == 4: # N or S
                    loop = 0
                    while loop < len(perspective_word):
                        temp_layer = persp_layer_names_list[loop]
                        pdb.gimp_image_set_active_layer(perspective_image, temp_layer)
                        if persp_vertical_justify == 2: # Right
                            word_x = persp_layer_names_list[0].width - temp_layer.width
                        elif persp_vertical_justify == 1: # centred
                            word_x = (persp_layer_names_list[0].width - temp_layer.width)/2
                        pdb.gimp_item_transform_translate(temp_layer, word_x, 0)
                        if scaled_scale>0: pdb.gimp_image_raise_item_to_top(perspective_image, temp_layer)
                        pdb.gimp_item_set_visible(temp_layer, True)
					
					
                        loop += 1
				
                pdb.gimp_image_resize_to_layers(perspective_image)
				
            if to_visible:
                to_vis_layer = pdb.gimp_layer_new_from_visible(perspective_image, perspective_image, "Combined layer")
                pdb.gimp_image_insert_layer(perspective_image, to_vis_layer, None, 0)
			
            #debug_string += " Current direction = " + str(direction) + " : offset_list = " + str(offset_list)
            #pdb.gimp_message(" Current direction = " + str(direction) + " left_perspective = " + str(left_perspective) )
			
			
            #pdb.gimp_message(str(debug_string) )
			
        else:
            message_text = "There appears to be no custom font layers associated with this image or the required character layers are missing from this file."
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message(message_text)
            pdb.gimp_message_set_handler(old_handler)    

    pdb.gimp_displays_flush()			
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
	

######################################################################################################################################
	
def proc_do_3D(image, style, scale_inc, depth, blur, use_decor, initial_shade, shade_range, darken_mode, direction, kerning_adjustment, do_preview, preview_char, do_perspective, persp_step, has_centre, left_perspective, right_perspective, persp_space_width, perspective_word, scaled_depth, scaled_scale, debug_string):


    test_for_custom_layer = False
		
    active_layer_id = pdb.gimp_image_get_active_layer(image)
    multi_layer_list = "No non-character layers."
	
    # set mode for 3D layers (increasingly darker)
    if darken_mode == 0:
        darken_mode = 35 # darken only
    elif darken_mode == 1:
        darken_mode = 54 # Luma/luminance darken
    elif darken_mode == 2:
        darken_mode = 30 # Multiply
    elif darken_mode == 3:
        darken_mode = 43 # Burn
    elif darken_mode == 4:
        darken_mode = 53 # Linear burn

    debug_string = debug_string
    vertical = False
    if direction == 0 or direction == 4: 
        vertical == True
	
    loop_ptr = 1
    first_char_layer_name = "none"
    num_layers, layer_ids = pdb.gimp_image_get_layers(image)
    loop = num_layers - loop_ptr
    layer_names_list = image.layers
    original_active_layer = image.active_layer
	
	
		
    num_multi_layers,other_layers,num_char_layers = 0,0,0
    visi_loop, first_char_layer  = 0, -1
    im_width = image.width
    im_height = image.height
	
	
    visi_loop = 0

    # loop through layers and turn off visibility of character layers
    while (visi_loop < num_layers and other_layers > 0):
        visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
        visi_name = visi_name.decode('utf-8')
        if visi_name == "space.png" or visi_name == " .png": 
            has_a_space = True
            space_loop = visi_loop
        if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
            # turn off visibility of letter layers
            pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)  
        visi_loop += 1
		
		
    preview_decorate_visibility = False

    visi_loop = 0
    while visi_loop < num_layers:
        # get info about layer we are working on
        layer_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
        char_layer_name = layer_name.decode('utf-8')
        # check to see if the current layer is a custom font character layer
        test_for_custom_layer = fn_test_for_custom_layer(image, layer_names_list[visi_loop], char_layer_name)
			
        if test_for_custom_layer:
            num_char_layers += 1
            if first_char_layer < 0: 
                first_char_layer = visi_loop
                first_char_layer_name = char_layer_name
                layer_names_list[visi_loop].set_offsets(0, 0)
				
        else:
            # look for the presence of decor layers - we use only the top one
            pdb.gimp_image_select_item(image, 2, layer_names_list[visi_loop])
            is_empty = pdb.gimp_selection_is_empty(image)
            if is_empty == False:
                if other_layers == 0:
                        
                    multi_layer_list = [layer_name]
                    preview_decorate_layer = layer_names_list[visi_loop]
                    preview_decorate_visibility = pdb.gimp_item_get_visible(layer_names_list[visi_loop])
                else:
                        
                    multi_layer_list.append(layer_name)
                other_layers += 1
				
            
        if not do_preview: pdb.gimp_item_set_visible(layer_names_list[visi_loop],False) # turn off visibility
        visi_loop += 1
		
    if str(multi_layer_list) == "No non-character layers.": 
        num_multi_layers = 0
        use_decor = False
    else:
        num_multi_layers = len(multi_layer_list)

    if do_preview:
        # preview character checks
        if preview_char == "": preview_char = "A"
        preview_char = preview_char + ".png"
	
        # check that the preview character is present or use another
        if preview_char in str(layer_names_list):
            prev_pos = str(layer_names_list).find(preview_char)
            prev_loop = str(layer_names_list).count(",",0,prev_pos)
        else:
            loop = 0
            preview_char = pdb.gimp_item_get_name(layer_names_list[loop])
            while preview_char=="space.png" or preview_char in multi_layer_list:
                loop += 1
                preview_char = pdb.gimp_item_get_name(layer_names_list[loop])

        preview_char = preview_char.decode('utf-8')	######## Added to catch higher value unicodes

			
    if not use_decor: num_multi_layers = 1 # ie we are going to copy the existing decorated characters
			
    if do_preview:
        # copy the preview character and paste it as a  new image
		
        preview_layer = pdb.gimp_image_get_layer_by_name(image, preview_char)
        success = pdb.gimp_edit_copy(preview_layer)
        preview_image = pdb.gimp_edit_paste_as_new_image(preview_layer)
        
        ##############################################################
        # begin undo for preview image
        pdb.gimp_image_undo_group_start(preview_image) ##############
		
        preview_display = pdb.gimp_display_new(preview_image)
        preview_num_layers, preview_layer_ids = pdb.gimp_image_get_layers(preview_image)
        preview_layer_names_list = preview_image.layers
		
		
	
    if num_multi_layers>0:
        visi_loop, max_width = 0, 0
        if use_decor: 
            decor_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[0])
            if do_preview:
                # copy decor layer from image and paste to preview image
                decor_copy = pdb.gimp_layer_new(preview_image, decor_layer.width, decor_layer.height, 1, "decor copy", 100, 0, )
                preview_image.add_layer(decor_copy, -1)
                success = pdb.gimp_edit_copy(decor_layer)
                # paste the selection into the floating layer
                floating_sel = pdb.gimp_edit_paste(decor_copy,True)
                # and anchor it
                pdb.gimp_floating_sel_anchor(floating_sel)
                # Re-position the copied layer
                decor_copy.set_offsets(0, 0)
				
        if do_preview:
            preview_loop = 0
            preview_im_width, preview_im_height = preview_image.width, preview_image.height
            while preview_loop < preview_num_layers:
                # get info about layer we are working on for custom font mode
                preview_layer_id = preview_layer_ids[preview_loop]
                preview_layer_name = preview_layer_names_list[preview_loop]

                # cater for unicode characters above normal range
                display_layer_name =  pdb.gimp_item_get_name(preview_layer_names_list[preview_loop])
                display_layer_name = display_layer_name.decode('utf-8')	
	
                # check we have the layer id of the currently active layer
                pdb.gimp_image_set_active_layer(preview_image, preview_layer_name)
                layer_id = pdb.gimp_image_get_active_layer(preview_image)

                # check to see if the current layer is a custom font character layer
                test_for_custom_layer = fn_test_for_custom_layer(preview_image, layer_id, display_layer_name)
	
                if test_for_custom_layer:
                    # ensure we have the active layer and make it visible
                    active_layer_id = pdb.gimp_image_get_active_layer(preview_image)
                    pdb.gimp_item_set_visible(active_layer_id, True)
			
                    # Take 3D direction from dialog
				
                    x_adj, y_adj = 1,1 # SE defaults
                    if direction == 0: x_adj, y_adj = 0,-1 # N
                    if direction == 1: x_adj, y_adj = 1,-1 # NE
                    if direction == 2: x_adj, y_adj = 1,0 # E
                    if direction == 3: x_adj, y_adj = 1,1 # SE
                    if direction == 4: x_adj, y_adj = 0,1 # S
                    if direction == 5: x_adj, y_adj = -1,1 # SW
                    if direction == 6: x_adj, y_adj = -1,0 # W
                    if direction == 7: x_adj, y_adj = -1,-1 # NW
				
                    # Step 1: Apply decor to initial character face if required by dialog
				
                    # make sure that existing layers are positioned at the origin of the image
                    preview_layer_names_list[preview_loop].set_offsets(0, 0)
					
                    # prepare to move layer to random position within the confines of the image size
                    layer_width = preview_layer_names_list[preview_loop].width
                    if layer_width > preview_im_width: layer_width = preview_im_width
                    layer_height = preview_layer_names_list[preview_loop].height
                    if layer_height > preview_im_height: layer_height = preview_im_height
					
                    # make a duplicate of the current character layer and insert
                    layer_position = pdb.gimp_image_get_item_position(preview_image, preview_layer_names_list[preview_loop])
                    layer_copy = pdb.gimp_layer_copy(preview_layer_names_list[preview_loop], True)
                    pdb.gimp_image_insert_layer(preview_image, layer_copy, None, layer_position)
					
					
				
                    if use_decor: preview_layer_names_list, layer_copy = proc_decor(preview_image, preview_im_width, layer_width, preview_im_height, layer_height, layer_copy, decor_layer, preview_layer_names_list, preview_loop)
					
                    # Step 2: Apply 3D effect to current character
					
					
                    # fill shade values taken from dialog
                    my_grey, shade_step = initial_shade, float(shade_range/depth)
                
                    grey_shade = gimpcolor.RGB(my_grey, my_grey, my_grey) # colour value
                    shade_loop_max, shade_loop  = depth, 0
                    decor_char = preview_layer_names_list[preview_loop] # get id of decorated character
                    scale_factor = scale_inc
				
                
                    if blur>0: 
                        # select the contents of the layer
                        pdb.gimp_image_select_item(preview_image, 2, layer_copy)
                        # shrink the selection
                        pdb.gimp_selection_shrink(preview_image, 2)
                        # invert selection
                        pdb.gimp_selection_invert(preview_image)
                        pdb.plug_in_gauss(preview_image, layer_copy, blur, blur, 0)
                        # cancel alpha selection
                        pdb.gimp_selection_none(preview_image)
						
						
                    if style == 0:
                        decor_char = proc_do_3D_translate(preview_image, decor_char, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_factor, blur, shade_loop, my_grey)
				
						
                    elif style == 1:
			
                        decor_char, layer_copy = proc_do_3D_scale(preview_image, decor_char, layer_copy, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_inc, blur, shade_loop, my_grey)
				
                        #pdb.gimp_message(err_list)
					
						
                    pdb.gimp_item_set_visible(decor_char, True)
				
                    if style ==  1 and (not use_decor): pdb.gimp_image_remove_layer(preview_image, layer_copy)
					
					
                    if use_decor: 					
                        pdb.gimp_image_remove_layer(preview_image, decor_copy)
						
                    preview_layer_names_list[preview_loop] = pdb.gimp_image_merge_visible_layers(preview_image, 0)
                    pdb.gimp_item_set_name(preview_layer_names_list[preview_loop], display_layer_name) # rename with character name
				
                    # note new width of character
                    new_width = preview_layer_names_list[preview_loop].width
                    if new_width>max_width: max_width = new_width
                    pdb.gimp_item_set_name(preview_layer_names_list[preview_loop], display_layer_name) # rename with character name
                    pdb.gimp_layer_resize(preview_layer_names_list[preview_loop],new_width,layer_height,0,0) # resize layer to new width
                    # Re-position the layer
                    preview_layer_names_list[preview_loop].set_offsets(0, 0)
                    pdb.gimp_item_set_visible(preview_layer_names_list[preview_loop], False)
			
			
                preview_loop +=1
        else:
		
		
            if do_perspective: # Set initial and final perspective direction values from dialog direction values
                
                if style == 0: # Translated
                    if direction == 0: # initial direction N mid-direction N final direction S
                        direction1, mid_direction, direction2, x_adj1, y_adj1, x_adj2, y_adj2 = 0, 0, 4, 0, -1, 0, 1
                    if direction == 1: # initial direction NE mid-direction N final direction NW
                        direction1, mid_direction, direction2, x_adj1, y_adj1, x_adj2, y_adj2 = 1, 0, 7, 1,-1, -1, -1
                    if direction == 2: # initial direction E mid-direction E final direction W
                        direction1, mid_direction, direction2, x_adj1, y_adj1, x_adj2, y_adj2 = 2, 2, 6, 1, 0, -1, 0
                    if direction == 3: # initial direction SE mid-direction S final direction SW
                        direction1, mid_direction, direction2, x_adj1, y_adj1, x_adj2, y_adj2 = 3, 4, 5, 1, 1, -1, 1
                    if direction == 4: # initial direction S mid-direction S final direction N
                        direction1, mid_direction, direction2, x_adj1, y_adj1, x_adj2, y_adj2 = 4, 4, 0, 0 ,1, 0, -1
                    if direction == 5: # initial direction SW mid-direction S final direction SE
                        direction1, mid_direction, direction2, x_adj1, y_adj1, x_adj2, y_adj2 = 5, 4, 3, -1, 1, 1, 1
                    if direction == 6: # initial direction W mid-direction W final direction E
                        direction1, mid_direction, direction2, x_adj1, y_adj1, x_adj2, y_adj2 = 6, 6, 2, -1, 0, 1, 0
                    if direction == 7: # initial direction NW mid-direction N final direction NE
                        direction1, mid_direction, direction2, x_adj1, y_adj1, x_adj2, y_adj2 = 7, 0, 1, -1, -1, 1, -1
		
            while visi_loop < num_layers:
                # get info about layer we are working on for custom font mode
                layer_id = layer_ids[visi_loop]
                layer_name = layer_names_list[visi_loop]

                # cater for unicode characters above normal range
                display_layer_name =  pdb.gimp_item_get_name(layer_names_list[visi_loop])
                display_layer_name = display_layer_name.decode('utf-8')	
	
                # check we have the layer id of the currently active layer
                pdb.gimp_image_set_active_layer(image, layer_name)
                layer_id = pdb.gimp_image_get_active_layer(image)

                # check to see if the current layer is a custom font character layer
                test_for_custom_layer = fn_test_for_custom_layer(image, layer_id, display_layer_name)
				
                if do_perspective: 
                    test_for_custom_layer = True # custom font layer names not used in perspective mode
	
                if do_perspective: 
                    debug_string += " depth at loop position " + str(visi_loop) + " = " + str(depth) + " : "
					
                if test_for_custom_layer:
                    # ensure we have the active layer and make it visible
                    active_layer_id = pdb.gimp_image_get_active_layer(image)
                    pdb.gimp_item_set_visible(active_layer_id, True)
					
			
                    # Take 3D direction from dialog
				
                    x_adj, y_adj = 1,1 # SE defaults
                    if direction == 0: x_adj, y_adj = 0,-1 # N
                    if direction == 1: x_adj, y_adj = 1,-1 # NE
                    if direction == 2: x_adj, y_adj = 1,0 # E
                    if direction == 3: x_adj, y_adj = 1,1 # SE
                    if direction == 4: x_adj, y_adj = 0,1 # S
                    if direction == 5: x_adj, y_adj = -1,1 # SW
                    if direction == 6: x_adj, y_adj = -1,0 # W
                    if direction == 7: x_adj, y_adj = -1,-1 # NW
				
                    # Step 1: Apply decor to initial character face if required by dialog
				
                    # make sure that existing layers are positioned at the origin of the image
                    layer_names_list[visi_loop].set_offsets(0, 0)
				
                    # prepare to move layer to random position within the confines of the image size
                    layer_width = layer_names_list[visi_loop].width
                    if layer_width > im_width: layer_width = im_width
                    layer_height = layer_names_list[visi_loop].height
                    if layer_height > im_height: layer_height = im_height
				
					
                    # make a duplicate of the current character layer and insert
                    layer_position = pdb.gimp_image_get_item_position(image, layer_names_list[visi_loop])
                    layer_copy = pdb.gimp_layer_copy(layer_names_list[visi_loop], True)
                    pdb.gimp_image_insert_layer(image, layer_copy, None, layer_position)
				
                    if use_decor: layer_names_list, layer_copy = proc_decor(image, im_width, layer_width, im_height, layer_height, layer_copy, decor_layer, layer_names_list, visi_loop)
					
                    # Step 2: Apply 3D effect to current character
                
                    # fill shade values taken from dialog
                    my_grey, shade_step = initial_shade, float(shade_range/depth)
                
                    grey_shade = gimpcolor.RGB(my_grey, my_grey, my_grey) # colour value
                    shade_loop_max, shade_loop  = depth, 0
                    decor_char = layer_names_list[visi_loop] # get id of decorated character
                    scale_factor = scale_inc
				
                
                    if blur>0: 
                        # select the contents of the layer
                        pdb.gimp_image_select_item(image, 2, layer_copy)
                        # shrink the selection
                        pdb.gimp_selection_shrink(image, 2)
                        # invert selection
                        pdb.gimp_selection_invert(image)
                        pdb.plug_in_gauss(image, layer_copy, blur, blur, 0)
                        # cancel alpha selection
                        pdb.gimp_selection_none(image)
						
                    verti_check = False
                    if vertical:
                        if has_centre:
							
                            if visi_loop == left_perspective + 1: verti_check = True
							
                        else:
						
                            if visi_loop == left_perspective: verti_check = True
                                
					
                    if (do_perspective and verti_check):
                        pdb.gimp_image_lower_item_to_bottom(image, layer_id)
						
                    
					
                    else:
				
                        if style == 0:
                            if "space" not in pdb.gimp_item_get_name(decor_char): # stop unnecessary processing for space character
                                decor_char = proc_do_3D_translate(image, decor_char, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_factor, blur, shade_loop, my_grey)
                            else:
                                pdb.gimp_image_lower_item_to_bottom(image, decor_char)
						
                        elif style == 1:
			
                            if "space" not in pdb.gimp_item_get_name(decor_char):
                                decor_char, layer_copy = proc_do_3D_scale(image, decor_char, layer_copy, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_inc, blur, shade_loop, my_grey)
                            else:
                                pdb.gimp_image_lower_item_to_bottom(image, decor_char)
					
						
                    pdb.gimp_item_set_visible(decor_char, True)
				
                    if style ==  1 and (not use_decor): pdb.gimp_image_remove_layer(image, layer_copy)
                
                    if use_decor: pdb.gimp_item_set_visible(decor_layer, False) # Turn off decor layer visibility
                    layer_names_list[visi_loop] = pdb.gimp_image_merge_visible_layers(image, 0)
                    pdb.gimp_item_set_name(layer_names_list[visi_loop], display_layer_name) # rename with character name
				
                    # note new width of character
                    new_width = layer_names_list[visi_loop].width
                    if new_width>max_width: max_width = new_width
                    pdb.gimp_item_set_name(layer_names_list[visi_loop], display_layer_name) # rename with character name
                    pdb.gimp_layer_resize(layer_names_list[visi_loop],new_width,layer_height,0,0) # resize layer to new width
                    # Re-position the layer
                    layer_names_list[visi_loop].set_offsets(0, 0)
                    pdb.gimp_item_set_visible(layer_names_list[visi_loop], False)
					
                    if do_perspective:
                        if style == 0: #Translated
                            if visi_loop<(left_perspective):
                                direction = direction1
                                depth -= persp_step
                            elif visi_loop==(left_perspective):
                                if has_centre:
                                    direction = direction2
                                    direction = mid_direction # Temp
                                else:
                                    direction = direction2
                            else:
                                direction = direction2
                                if has_centre:
                                    if visi_loop==(left_perspective+1):
                                        dummy = 0
                                    else:
                                        depth += persp_step
						
                                else:
                                    depth += persp_step
                        if style == 1: # Scaled
                            depth += scaled_depth
                            #pdb.gimp_message("depth = " + str(depth) + " scaled_depth = " + str(scaled_depth) )
                visi_loop += 1
							
				
			
        if use_decor: 
            if not do_preview: pdb.gimp_item_set_visible(decor_layer, False) # Turn off decor layer visibility
			
        if do_preview: 
            pdb.gimp_image_resize(preview_image, max_width, im_height, 0, 0) # resize canvas to fit characters
            preview_layer = pdb.gimp_image_get_layer_by_name(preview_image, preview_char)
            pdb.gimp_item_set_visible(preview_layer, True)
            if use_decor: pdb.gimp_item_set_visible(preview_decorate_layer,  preview_decorate_visibility)
			
            ###########################################
            # Set end undo point for preview image
            pdb.gimp_image_undo_group_end(preview_image)
            ###########################################
		
        if not do_perspective and (first_char_layer_name <> "none"): # the character layer to make visible on completion
            first_char_layer_id = pdb.gimp_image_get_layer_by_name(image, first_char_layer_name)
            pdb.gimp_item_set_visible(first_char_layer_id, True)
            pdb.gimp_image_set_active_layer(image,first_char_layer_id)
				
			
    return max_width, im_height, debug_string
			
######################################################################################################################################

def proc_decor(image, im_width, layer_width, im_height, layer_height, layer_copy, decor_layer, layer_names_list, visi_loop):

    # set random position for copying
    random.seed()
    random_x = random.randrange(0, (im_width + 1) - layer_width)
    random.seed()
    random_y = random.randrange(0, (im_height + 1) - layer_height)

    # set random position to copy visible layers 
    layer_copy.set_offsets(random_x, random_y)
			
			
    # select the contents of the layer and clear contents
    pdb.gimp_image_select_item(image, 2, layer_copy)
    pdb.gimp_drawable_edit_clear(layer_copy)
			
    # copy the contents of the visible layers into the selection
    pdb.gimp_item_set_visible(decor_layer, True)
    pdb.gimp_selection_grow(image, 1)
    success = pdb.gimp_edit_copy_visible(image)
				
    # turn on layer visibility
    pdb.gimp_item_set_visible(layer_copy, True) 
					
    # paste the selection into the floating layer
    floating_sel = pdb.gimp_edit_paste(layer_copy,True)
    # and anchor it
    pdb.gimp_floating_sel_anchor(floating_sel)
			
    # Re-position the copied layer
    layer_copy.set_offsets(0, 0)
				
    # Make the original character layer visible before we merge down					
    pdb.gimp_item_set_visible( layer_names_list[visi_loop], True)
				
    # cancel alpha selection
    pdb.gimp_selection_none(image)
    # clear the character layer
    pdb.gimp_drawable_edit_clear(layer_names_list[visi_loop])
				
                   
    # Merge down the copied layer to the character layer
    layer_copy = pdb.gimp_image_merge_down(image, layer_copy, 0)
    # Re-set the id of the layer
    layer_names_list[visi_loop] = pdb.gimp_image_get_active_layer(image)
				
    # Turn off layer visibility
    pdb.gimp_item_set_visible(layer_names_list[visi_loop], False) 
	
    return layer_names_list, layer_copy
	
	

######################################################################################################################################


def proc_do_3D_translate(image, decor_char, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_factor, blur, shade_loop, my_grey):

    while shade_loop<shade_loop_max:
        new_grey = (1.0/255) * my_grey
        grey_shade = gimpcolor.RGB(new_grey, new_grey, new_grey)
        pdb.gimp_context_set_foreground(grey_shade)
                    
        if shade_loop>0:
            decor_char_position = pdb.gimp_image_get_item_position(image, decor_char) # find position of layer in the item tree
            decor_char_copy2 = pdb.gimp_layer_copy(decor_char, True) # copy and insert layer
            pdb.gimp_image_insert_layer(image, decor_char_copy2, None, decor_char_position)
            fill_layer2 = gimp.Layer(image, "Fill Layer2", decor_char.width, decor_char.height, RGBA_IMAGE, 100, darken_mode) # add a fill layer with selected darker mode
            pdb.gimp_drawable_fill(fill_layer2, 0) # fill layer with current shade of grey
            image.add_layer(fill_layer2, -1)
            pdb.gimp_item_set_visible(decor_char_copy2, True) # Make sure decor layer copy is visible
            fill_layer2 = pdb.gimp_image_merge_down(image, fill_layer2, 2) #  and merge down the fill layer onto the decor layer copy
            pdb.gimp_item_set_name(fill_layer2, "Fill Layer2")
            pdb.gimp_layer_set_mode(fill_layer2, 28) # set mode to normal
            pdb.gimp_layer_translate(fill_layer2, ((0+shade_loop)* x_adj)*scale_factor, ((0+shade_loop)*y_adj)*(scale_factor/2))
            # move the layer by 3D amount
            pdb.gimp_image_lower_item_to_bottom(image, fill_layer2)
            if blur>0: pdb.plug_in_gauss(image, fill_layer2, blur, blur, 0)
            if (fill_layer2.width<4) or  (fill_layer2.height<4): shade_loop=shade_loop_max # abort loop before layer gets too small
        else:
            # first loop specifics
            decor_char_position = pdb.gimp_image_get_item_position(image, decor_char) # find position of layer in the item tree
            decor_char_copy = pdb.gimp_layer_copy(decor_char, True) # copy and insert layer
            pdb.gimp_image_insert_layer(image, decor_char_copy, None, decor_char_position)
            fill_layer = gimp.Layer(image, "Fill Layer", decor_char_copy.width, decor_char_copy.height, RGBA_IMAGE, 100, darken_mode) # add a fill layer with selected darker mode
            pdb.gimp_drawable_fill(fill_layer, 0) # fill layer with current shade of grey
            image.add_layer(fill_layer, -1)
            pdb.gimp_item_set_visible(decor_char_copy, True) # Make sure decor layer copy is visible
            fill_layer = pdb.gimp_image_merge_down(image, fill_layer, 2) #  and merge down the fill layer onto the decor layer copy
            pdb.gimp_item_set_name(fill_layer, "Fill Layer")
            pdb.gimp_layer_set_mode(fill_layer, 28) # set mode to normal
            pdb.gimp_layer_translate(fill_layer, x_adj, y_adj) # move the layer
            pdb.gimp_image_lower_item_to_bottom(image, fill_layer)
            if blur>0: pdb.plug_in_gauss(image, fill_layer, blur, blur, 0)
        my_grey -= shade_step # adjust grey shade value
					
        shade_loop += 1
		
    return decor_char

######################################################################################################################################
	
def proc_do_3D_scale(image, decor_char, layer_copy, x_adj, y_adj, shade_loop_max, darken_mode, shade_step, scale_inc, blur, shade_loop, my_grey):

    scale_factor = scale_inc
    translate_inc = scale_factor
    while shade_loop<shade_loop_max:
        new_grey = (1.0/255) * my_grey
        grey_shade = gimpcolor.RGB(new_grey, new_grey, new_grey)
        pdb.gimp_context_set_foreground(grey_shade)
        if shade_loop>0:
                            
            decor_char_position = pdb.gimp_image_get_item_position(image, decor_char) # find position of layer in the item tree
            decor_char_copy2 = pdb.gimp_layer_copy(decor_char, True) # copy and insert layer
            pdb.gimp_image_insert_layer(image, decor_char_copy2, None, decor_char_position)
            fill_layer2 = gimp.Layer(image, "Fill Layer2", decor_char.width, decor_char.height, RGBA_IMAGE, 100, darken_mode) # add a fill layer with selected darker mode
            pdb.gimp_drawable_fill(fill_layer2, 0) # fill layer with current shade of grey
            image.add_layer(fill_layer2, -1)
            pdb.gimp_item_set_visible(decor_char_copy2, True) # Make sure decor layer copy is visible
            fill_layer2 = pdb.gimp_image_merge_down(image, fill_layer2, 2) #  and merge down the fill layer onto the decor layer copy
            pdb.gimp_item_set_name(fill_layer2, "Fill Layer2")
            pdb.gimp_layer_set_mode(fill_layer2, 28) # set mode to normal
            pdb.gimp_layer_scale(fill_layer2,layer_copy.width-scale_factor, layer_copy.height-scale_factor, False) # scale the current layer by the requested amount
							
            pdb.gimp_item_transform_translate(fill_layer2, x_adj*scale_factor, y_adj * (scale_factor/2)) # move the layer
            pdb.gimp_image_lower_item_to_bottom(image, fill_layer2)
            if blur>0: pdb.plug_in_gauss(image, fill_layer2, blur, blur, 0)
            err_list.append("scale_factor = " + str(scale_factor) + " : x_adj =" + str((1+shade_loop)* x_adj) )
            if (fill_layer2.width<4) or  (fill_layer2.height<4): shade_loop=shade_loop_max # abort loop before layer gets too small
        else:
            # first loop specifics
            pdb.gimp_item_set_visible(layer_copy, False)
            decor_char_position = pdb.gimp_image_get_item_position(image, decor_char) # find position of layer in the item tree
            decor_char_copy = pdb.gimp_layer_copy(decor_char, True) # copy and insert layer
            pdb.gimp_image_insert_layer(image, decor_char_copy, None, decor_char_position)
            fill_layer = gimp.Layer(image, "Fill Layer", decor_char_copy.width, decor_char_copy.height, RGBA_IMAGE, 100, darken_mode) # add a fill layer with selected darker mode
            pdb.gimp_drawable_fill(fill_layer, 0) # fill layer with current shade of grey
            image.add_layer(fill_layer, -1)
            pdb.gimp_item_set_visible(decor_char_copy, True) # Make sure decor layer copy is visible
            fill_layer = pdb.gimp_image_merge_down(image, fill_layer, 2) #  and merge down the fill layer onto the decor layer copy
            pdb.gimp_item_set_name(fill_layer, "Fill Layer")
            pdb.gimp_layer_set_mode(fill_layer, 28) # set mode to normal
							
            pdb.gimp_layer_scale(fill_layer,fill_layer.width-scale_factor, fill_layer.height-scale_factor, False) # scale first layer
            pdb.gimp_item_transform_translate(fill_layer, x_adj*scale_factor, y_adj * (scale_factor/2)) # move the layer
            pdb.gimp_image_lower_item_to_bottom(image, fill_layer)
            if blur>0: pdb.plug_in_gauss(image, fill_layer, blur, blur, 0)
            err_list = ["scale_factor = " + str(scale_factor) + " : x_adj =" + str(x_adj)]
        my_grey -= shade_step # adjust grey shade value
        scale_factor += scale_inc
        translate_inc += scale_inc
        
					
        shade_loop += 1
		
		
    return decor_char, layer_copy
	
    #### Useful Functions
# -------------------------------------------------------------------------------------
# -------------------------------------------------------------------------------------

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value


######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################
		
def proc_trim_away(img, layer, amount):
    #### Trim off pixels from a layer
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
    #### End of Proc_trim_away
	

######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Faux_3D_v0-04_Tabbed",
    "Adds 3D effect to custom font characters",
    "Adds 3D effect to custom font characters",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2022",
    "<Image>/Filters/Custom Font Tools/Custom Font Faux 3D v0.04 Tabbed...",
    "RGB*",
    [
    (PF_OPTION, "style_option", "3D style", 0, ("Style 1 - Translated", "Style 2 - Scaled"),{'newLine':'1','tab':0}),
    (PF_SPINNER, "scale_inc_value", "Style scale factor: ", 2, (2, 20, 1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "depth", "3D depth: ", 8, (2, 400, 1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "blur", "3D layer blur: ", 4, (1, 20, 1),{'newLine':'1','tab':0}),
    (PF_TOGGLE, "pf_Use_decor", "Use image as decor face:", 0,{'newLine':'1','tab':0}),
    (PF_SPINNER, "initial_shade", "Initial shade for depth: ", 250, (64,250,1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "final_shade", "Range for shade: ", 96, (64,96,1),{'newLine':'1','tab':0}),
    (PF_OPTION, "mode_option", "Darken mode used: ", 4, ("Darken only", "Luma darken", "Multiply", "Burn", "Linear burn"),{'newLine':'1','tab':0}),
    (PF_OPTION, "option", "3D direction", 3, ("North", "North-East", "East", "South-East","South","South-West","West","North-West"),{'newLine':'1','tab':0}),
    (PF_TOGGLE, "adjust_kerning", "Adjust kerning paths: ", 0,{'newLine':'1','tab':0}),
    (PF_TOGGLE, "do_perspective", "Create word using perspective : ", 0,[("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_STRING, "perspective_string", "Perspective word: ", "GIMP", 0,{'newLine':'1','tab':1}),
    (PF_SPINNER, "perspective_spacing", "Adjust spacing between 3D characters: ", 0, (-100,100,1),{'newLine':'1','tab':1}),
    (PF_OPTION, "justify_option", "Vertical perspective justification: ", 0, ("To left", "To centre", "To right"),{'newLine':'1','tab':1}),
    (PF_OPTION, "justify_option2", "Horizontal perspective justification: ", 2, ("To top", "To centre", "To bottom"),{'newLine':'1','tab':1}),
    (PF_SPINNER, "scaled_depth", "Depth step (Scaled style only): ", 0, (-100,100,1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "scaled_scale", "Scaling step (Scaled style only): ", 0, (-100,200,1),{'newLine':'1','tab':1}),
    (PF_TOGGLE, "do_visible", "Add a merged view layer: ", 1,[("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "do_preview", "Preview single character\n(to a new image): ", 0,[("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_STRING, "preview_string", "Preview character: ", "A", 0,{'newLine':'1','tab':1}),
    (PF_TAB    , "ptab"    , "tabs", 0, (_('Custom Font Options'),_('3D word & Preview Options') ) ),
    ],
    [],
    do_stuff)

main()
